from .plot import (create_pie,
                   create_box,
                   create_scatter,
                   create_treemap,
                   create_violin,
                   create_sunburst,
                   create_correlation_plot,
                   create_histogram)
from .utils import (columns_with_distinct_values,
                    extract_col_types,
                    figures_to_html)
import pandas as pd


class AutoPlot(object):
    def __init__(self,
                 df,
                 nbins=30,
                 scatter_maximum_color_groups=5,
                 marginal_x=None,
                 marginal_y=None,
                 cumulative=False,
                 filename='autoplot',
                 box_points='outliers',
                 boxmode='group',
                 box_notched=False,
                 log_x=False,
                 log_y=False,
                 size=None,
                 histfunc='count',
                 histnorm=None,
                 include_scatter=False,
                 include_box=True,
                 include_treemap=True,
                 include_sunburst=True,
                 include_correlation=False,
                 include_pie=True,
                 include_histogram=True,
                 include_violin=True,
                 violinmode='group',
                 violin_box=True,
                 violin_points='all',
                 treemap_path_limit=2,
                 sunburst_path_limit=2,
                 correlation_method='pearson',
                 maximum_number_sectors=3,
                 maximum_number_boxplots=5,
                 maximum_number_violinplots=5,
                 pie_chart_hole=False,
                 create_html=True,
                 scatter_orientation='v',
                 opacity = 1.0
                 ):
        """
        Entry point for automated data visualization
        :param df:
        :param nbins:
        :param marginals:
        :param cumulative:
        :param filename:
        :param box_points:
        :param boxmode:
        :param box_notched:
        :param log_x:
        :param log_y:
        :param size:
        :param histfunc:
        :param histnorm:
        :param include_scatter: boolean, whether to include scatter plots
        :param include_box: boolean, whether to include box plots
        :param include_treemap: boolean, whether to include treemaps
        :param include_sunburst: boolean, whether to include sunburst
        :param include_correlation: boolean, whether to include correlation plot
        :param include_pie: boolean, whether to include pie
        :param include_histogram: boolean, whether to include histogram
        :param include_violin: boolean, whether to include violin plots
        :param violinmode: str, display mode for violin charts
        :param violin_box: boolean, whether to include box plot in violin
        :param violin_points: str, how to display points in a violin plot
        :param treemap_path_limit:
        :param sunburst_path_limit:
        :param correlation_method: str, method to use to compute correlation
        :param maximum_number_sectors: int, maximum number of sectors in pie charts
        :param maximum_number_boxplots:int, maximum_number_boxplots
        :param pie_chart_hole: boolean, whether to include a hole in pie chart or not
        :param create_html:boolean, whether to create an html file or not

        Usage
        >>> import pandas as pd
        >>> from datamallet.visualization import AutoPlot
        >>> df3 = pd.DataFrame({'A':[1,1,2,1,1], 'B':[2,2,1,2,0],})
        >>> autoplot1 = AutoPlot(df=df3,filename='autoplot' ,create_html=True,include_scatter=True)
        >>> print(autoplot1.chart_type()) # output may change as more chart types are added

        ['scatter', 'correlation_plot', 'histogram', 'boxplot', 'violinplot']

        >>> autoplot1.show()
        # creates a file autoplot.html based on the filename you provide, all charts would be found there

        """
        self.df = df
        self.nbins = nbins
        self.scatter_maximum_color_groups = scatter_maximum_color_groups
        self.scatter_orientation = scatter_orientation
        self.opacity=opacity
        self.marginal_x = marginal_x
        self.marginal_y = marginal_y
        self.cumulative = cumulative
        self.filename = filename
        self.points = box_points
        self.boxmode = boxmode
        self.notched = box_notched
        self.log_x = log_x
        self.log_y = log_y
        self.size = size
        self.histfunc = histfunc
        self.histnorm = histnorm
        self.column_types = extract_col_types(df=df)
        self.include_scatter = include_scatter
        self.include_box = include_box
        self.include_treemap = include_treemap
        self.include_sunburst = include_sunburst
        self.include_correlation = include_correlation
        self.include_pie = include_pie
        self.include_histogram = include_histogram
        self.include_violin = include_violin
        self.violinmode = violinmode
        self.violin_box = violin_box
        self.violin_points = violin_points
        self.treemap_path_limit = treemap_path_limit
        self.sunburst_path_limit = sunburst_path_limit
        self.correlation_method = correlation_method
        self.maximum_number_sectors = maximum_number_sectors
        self.maximum_number_boxplots = maximum_number_boxplots
        self.maximum_number_violinplots = maximum_number_violinplots
        self.create_html = create_html
        self.pie_chart_hole = pie_chart_hole
        self.pie_sectors = columns_with_distinct_values(df=self.df, categorical_only=True,
                                                        maximum_number_distinct_values=self.maximum_number_sectors)
        assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
        assert isinstance(nbins, int), "nbins must be an integer"
        assert isinstance(scatter_maximum_color_groups, int)
        assert marginal_x in ['rug', 'box', 'violin', 'histogram', None]
        assert marginal_y in ['rug', 'box', 'violin', 'histogram', None]
        assert isinstance(scatter_orientation, str)
        assert scatter_orientation in ['v', 'h']
        assert isinstance(cumulative, bool), "cumulative must be a boolean"
        assert '.' not in filename, "filename doesn't need an extension"
        assert isinstance(filename, str), "filename must be a string with a dot or an extension"
        assert box_points in ['all', 'outliers', 'suspectedoutliers', False], "accepted values for points 'all', " \
                                                                          "'outliers', " \
                                                                          "'suspectedoutliers'"
        assert isinstance(boxmode, str), "boxmode must be a string"
        assert boxmode in ['group', 'overlay'], "boxmode must be either group or overlay"
        assert isinstance(box_notched, bool), "notched must be a boolean"
        assert isinstance(log_y, bool),"log_x must be a boolean"
        assert isinstance(log_x, bool),"log_y must be a boolean"
        assert isinstance(size, str) or size is None
        assert histnorm in ['percent', 'probability', 'density', 'probability density', None]
        assert histfunc in ['count', 'sum', 'avg', 'min',
                            'max'], "histfunc must be one of 'count', 'sum', 'avg','min','max' "
        assert violinmode in ['group', 'overlay'], "violinmode must be either group or overlay"
        assert isinstance(violin_box, bool)
        assert violin_points in ['all', 'outliers', 'suspectedoutliers', False]
        assert treemap_path_limit > 1
        assert sunburst_path_limit > 1
        assert correlation_method in ['pearson', 'kendall', 'spearman']
        assert maximum_number_sectors > 1
        assert isinstance(maximum_number_sectors,int)
        assert isinstance(maximum_number_boxplots,int)
        assert isinstance(maximum_number_violinplots,int)
        assert isinstance(pie_chart_hole, bool)
        assert isinstance(opacity, float)
        assert opacity <= 1.0
        assert isinstance(include_pie, bool)
        assert isinstance(include_treemap,bool)
        assert isinstance(include_box, bool)
        assert isinstance(include_correlation, bool)
        assert isinstance(include_histogram, bool)
        assert isinstance(include_scatter,bool)
        assert isinstance(include_sunburst,bool)
        assert isinstance(include_violin, bool)
        assert isinstance(create_html,bool)
        # assert marginal_x in ['rug', 'box', 'violin', 'histogram', None]
        # assert marginal_y in ['rug', 'box', 'violin', 'histogram', None]
        # assert isinstance(cumulative, bool), "cumulative must be a boolean"

    def chart_type(self):
        column_types = self.column_types
        numeric_cols = column_types['numeric']
        datetime_cols = column_types['datetime']
        timedelta_cols = column_types['timedelta']
        categorical_cols = column_types['categorical']
        boolean_cols = column_types['boolean']
        object_cols = column_types['object']

        chart_types = list()

        if len(numeric_cols) == 0:
            # if no numeric columns,no charts can be made, return empty list
            return chart_types

        if len(numeric_cols) > 1:
            chart_types.append('scatter')
            chart_types.append('correlation_plot')
        if len(numeric_cols) > 0:
            chart_types.append('histogram')
            chart_types.append('boxplot')
            chart_types.append('violinplot')
        if len(categorical_cols) >0 or len(boolean_cols) > 0 or len(object_cols):
            chart_types.append('treemaps')
            chart_types.append('sunburst')
        if len(datetime_cols) !=0 or len(timedelta_cols) != 0:
            chart_types.append('timeseries_plot')
        if len(self.pie_sectors) !=0:
            chart_types.append('pie')

        return chart_types

    def show(self):
        chart_types = self.chart_type()

        object_cols = self.column_types['object']
        boolean_cols = self.column_types['boolean']
        categorical_cols = self.column_types['categorical']
        numeric_cols = self.column_types['numeric']

        categorical = list()
        categorical.extend(object_cols)
        categorical.extend(boolean_cols)
        categorical.extend(categorical_cols)

        figure_list = list()

        if len(chart_types) != 0:
            for chart in chart_types:
                if chart == 'pie' and self.include_pie:

                    pie_list = create_pie(df=self.df,
                                          numeric_cols=numeric_cols,
                                          list_of_categorical_columns=self.pie_sectors,
                                          create_html=False,
                                          hole=self.pie_chart_hole,
                                          filename='pie')

                    figure_list.extend(pie_list)
                if chart == 'scatter' and self.include_scatter:
                    # if len(categorical) == 0:

                    scatter_plot_list = create_scatter(df=self.df,
                                                       col_types= self.column_types,
                                                       marginal_y=self.marginal_x,
                                                       marginal_x=self.marginal_y,
                                                       orientation=self.scatter_orientation,
                                                       opacity=self.opacity,
                                                       maximum_color_groups=self.scatter_maximum_color_groups,
                                                       log_x=self.log_x,
                                                       log_y=self.log_y,
                                                       create_html=False)

                    figure_list.extend(scatter_plot_list)

                if chart == 'correlation_plot' and self.include_correlation:
                    correlation_plot_list = create_correlation_plot(df=self.df,
                                                                    correlation_method=self.correlation_method)
                    figure_list.extend(correlation_plot_list)

                if chart == 'histogram' and self.include_histogram:
                    histogram_list = create_histogram(df=self.df,
                                                      numeric_cols=numeric_cols,
                                                      nbins=self.nbins,
                                                      marginal=self.marginal_x,
                                                      cumulative=self.cumulative,
                                                      histfunc=self.histfunc,
                                                      histnorm=self.histnorm,
                                                      filename='histogram',
                                                      create_html=False)

                    figure_list.extend(histogram_list)

                if chart == 'boxplot' and self.include_box:
                    box_list = create_box(df=self.df,
                                          maximum_number_boxplots=self.maximum_number_boxplots,
                                          col_types= self.column_types,
                                          points=self.points,
                                          boxmode=self.boxmode,
                                          notched=self.notched,
                                          color=None,
                                          filename='box',
                                          create_html=False)

                    figure_list.extend(box_list)

                if chart == 'treemaps' and self.include_treemap:
                    treemap_list = create_treemap(df=self.df,
                                                  col_types=self.column_types,
                                                  create_html=False,
                                                  filename='treemap',
                                                  limit=self.treemap_path_limit)
                    figure_list.extend(treemap_list)

                if chart == 'sunburst' and self.include_sunburst:
                    sunburst_list = create_sunburst(df=self.df,
                                                    col_types=self.column_types,
                                                    create_html=False,
                                                    filename='sunburst',
                                                    limit=self.sunburst_path_limit)

                    figure_list.extend(sunburst_list)

                if chart == 'violinplot' and self.include_violin:
                    violin_list = create_violin(df=self.df,
                                                col_types= self.column_types,
                                                maximum_number_violinplots=self.maximum_number_violinplots,
                                                filename='violin',
                                                create_html=False,
                                                points=self.violin_points,
                                                display_box=self.violin_box,
                                                color=None)

                    figure_list.extend(violin_list)

        if self.create_html:
            figures_to_html(figs=figure_list, filename=self.filename)

        return figure_list



from datamallet.tabular.utils import (check_columns,
                                      extract_col_types,
                                      get_unique,
                                      check_numeric,
                                      unique_count, combine_categorical_columns)
import pandas as pd


def columns_with_distinct_values(df,
                                 maximum_number_distinct_values=3,
                                 categorical_only=True):
    """
    Determines columns in a dataframe whose unique value count is
    less than or equal to the maximum_number_distinct_values
    :param df: pandas dataframe
    :param maximum_number_distinct_values:int, the upper limit on
            the number of distinct values in a column (translates into number of sectors in pie chart).
            A value of 3 is good for pie charts, 7 for boxplots or violin plots.
    :param categorical_only: boolean, whether to include categorical columns only in the final list or not
    :return: a list of column names which conform to the columns
                which have the number of distinct values less than the specified maximum_number_distinct_values
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(maximum_number_distinct_values, int), "maximum_number_sectors must be an integer"
    assert isinstance(categorical_only, bool), "category_only must be a boolean"

    columns_list = list()
    unique_count_dict = unique_count(df=df) # dictionary of column name with count of unique values as key

    col_types = extract_col_types(df=df)
    if categorical_only:
        categorical_cols = combine_categorical_columns(df, col_types)
        for col in df.columns:
            if col not in categorical_cols:
                del unique_count_dict[col] # delete non categorical columns

    for column_name, distinct_count in unique_count_dict.items():
        if distinct_count <= maximum_number_distinct_values:
            columns_list.append(column_name)

    return columns_list


def column_use(df,
               col_types,
               threshold=5):
    """
    This function helps in determining whether a column in a dataframe
       should be used to color the data points in a chart or if it is
        okay for simply naming the points
    :param df: pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param threshold: int, think of it as the number
           of distinct colors in a chart(e.g scatterplot)
    :return: a dictionary that decides what columns to be used to color points in a chart,
            and which should be used to name points, they keys are name and hue, values are list of column names
    """
    assert isinstance(col_types, dict), "col_types must be a dictionary"
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(threshold, int), "threshold must be an integer"
    assert 'numeric' in col_types.keys(), "col_types dictionary missing key numeric"
    assert 'object' in col_types.keys(), "col_types dictionary missing key object"
    assert 'boolean' in col_types.keys(), "col_types dictionary missing key boolean"
    assert 'categorical' in col_types.keys(), "col_types dictionary missing key categorical"
    assert 'datetime' in col_types.keys(), "col_types dictionary missing key datetime"
    assert 'timedelta' in col_types.keys(), "col_types dictionary missing key timedelta"
    all_categorical_cols = combine_categorical_columns(df=df, col_types=col_types)

    number_of_rows = df.shape[0]

    column_use_dict = dict()
    names_list = list()
    hue_list = list()

    for col in all_categorical_cols:
        number_of_unique = len(get_unique(df=df, col_name=col))
        ratio = number_of_rows / number_of_unique

        if ratio < threshold:
            names_list.append(col)

        else:
            hue_list.append(col)

    column_use_dict['name'] = names_list
    column_use_dict['hue'] = hue_list

    return column_use_dict


def figures_to_html(figs, filename):
    """
    Utility function that creates a single html file when given the list of plotly graph objects
    :param figs: python list of plotly graph objects (plotly chart objects).
    :param filename: str, string nmae for file excluding any extension
    :return: None
    """
    assert isinstance(figs, list), "figs is expected to be a list of plotly graph objects"
    assert isinstance(filename, str), "filename must be a string"
    assert '.' not in filename, "no dot should be in filename"
    html_filename = filename+'.html'

    with open(html_filename, 'a') as f:
        for fig in figs:
            f.write(fig.to_html(full_html=False, include_plotlyjs='cdn'))

    return None


def create_pairs(df, numeric_cols):
    """
    Create a non repeat list of tuples which contains pairs of
    numeric cols in df for visualization purpose as x and y axis
    :param df: pandas dataframe
    :param numeric_cols: list of column names which have numeric data, output of extract_numeric_cols(df=df)
    :return: list of tuple of non repeat pairing of columns in numeric_cols
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(numeric_cols, list), "numeric_cols must be a list"
    assert len(numeric_cols) >= 2, "the length of numeric_cols must be greater than 2"

    passed_cols = set()
    pairs = list()

    if check_columns(df=df,column_list=numeric_cols) and check_numeric(df=df,column_list=numeric_cols):
        for col1 in numeric_cols:
            for col2 in numeric_cols:
                if col1 == col2:
                    continue

                if len(passed_cols) == 0 or col2 not in passed_cols:
                    pairs.append((col1,col2))

            passed_cols.add(col1)

    return pairs


def hierarchical_path(df,col_types, limit=3):
    """
    It helps to determine the path for a tree map or sunburst chart (hierarchical charts),
    the idea is to start the path from the column with the
    least number of unique values to the one with the most
    :param df: pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param limit: the number of elements in the path list,
                i.e maximum number of treemap/sunburst categories
    :return:sorted_cols: list of column names which are categorical in nature,
            starting with the columns with least number of unique to the most.
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(limit, int), "limit must be a int"
    assert limit >1, "limit must be at least 2, to have a meaningful hierarchical chart"
    assert isinstance(col_types, dict), "col_types must be a dictionary"
    assert 'numeric' in col_types.keys(), "col_types dictionary missing key numeric"
    assert 'object' in col_types.keys(), "col_types dictionary missing key object"
    assert 'boolean' in col_types.keys(), "col_types dictionary missing key boolean"
    assert 'categorical' in col_types.keys(), "col_types dictionary missing key categorical"
    assert 'datetime' in col_types.keys(), "col_types dictionary missing key datetime"
    assert 'timedelta' in col_types.keys(), "col_types dictionary missing key timedelta"

    col_list = list()
    sorted_cols = list()

    col_list.extend(col_types['object'])
    col_list.extend(col_types['boolean'])
    col_list.extend(col_types['categorical'])

    if len(col_list) >= 1:
        if len(col_list) == 1:
            # no need to sort if there is only one element in the list
            return col_list
        else:
            # returns a count of unique values in each column
            unique_counts = [len(get_unique(df=df,col_name=x)) for x in col_list]
            sorted_columns = [col_name for _,col_name in sorted(zip(unique_counts, col_list))]
            sorted_cols = sorted_columns[:limit]

    return sorted_cols
















from .utils import (hierarchical_path, columns_with_distinct_values,
                    create_pairs,
                    figures_to_html)
from datamallet.tabular.utils import (check_categorical,
                                      check_numeric,
                                      extract_numeric_cols,
                                      calculate_correlation)
import pandas as pd
import plotly.express as px


def create_pie(df,
               numeric_cols,
               list_of_categorical_columns,
               create_html=False,
               hole=False,
               filename='pie'):
    """
    Creates a pie chart for every categorical variable in the dataset.
    :param df: pandas dataframe,
    :param numeric_cols: list of columns with numeric data
    :param list_of_categorical_columns:list of column names which have categorical data.
            Output of function columns_with_distinct_values(df=self.df, categorical_only=True,
                                                        maximum_number_distinct_values=maximum_number_sectors)
            Also output of extract_categorical_cols in tabular/utils module
    :param create_html:boolean, whether the figures should be converted to HTML or not
    :param hole:boolean, hole in the pie chart
    :param filename:str, a suitable name for the produced html file, exclude the extension
    :return: list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(numeric_cols, list), "numeric_cols must be a list"
    assert isinstance(list_of_categorical_columns, list), "list_of_categorical_columns must be a list"
    assert len(list_of_categorical_columns) != 0, "list_of_categorical_columns must not be empty"
    assert len(numeric_cols) != 0, "numeric_cols must not be empty"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert isinstance(hole, bool), "hole must be a boolean"
    assert isinstance(filename, str), "filename must be a string without a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"

    figure_list = list()

    if check_numeric(df=df,column_list=numeric_cols) and check_categorical(df=df,
                                                                           column_list=list_of_categorical_columns):

        for value in numeric_cols:
            for name in list_of_categorical_columns:
                plot = px.pie(names=name,
                              data_frame=df,
                              values=value,
                              hole=hole,
                              title='Pie chart showing distribution of {} across {} segments'.format(value, name))

                figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_violin(df,
                  col_types,
                  filename='violin',
                  create_html=True,
                  violinmode='group',
                  points='all',
                  display_box=True,
                  color=None,
                  maximum_number_violinplots=7
                  ):
    """
    Creates a list of violin plot figure objects
    :param df: pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param filename: str, filename for the html file
    :param create_html: boolean, whether to create an html file or not
    :param violinmode: str, how you want the charts to be displayed
    :param points: str, how the points in the violin chart should be displayed
    :param display_box: boolean, whether to display a box within the violin chart
    :param color:
    :param maximum_number_violinplots:
    :return:list which contains plotly graph objects
    """
    assert isinstance(df,pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(col_types, dict), "col_types must be a dictionary"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert isinstance(violinmode,str), "violinmode must be a string"
    assert isinstance(points,str), "points must be a string"
    assert isinstance(display_box,bool), "display_box must be a boolean"
    assert isinstance(maximum_number_violinplots, int), "maximum_number_violinplots must be an integer"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert violinmode in ['group', 'overlay'],"violinmode must be either group or overlay"
    assert points in ['all', 'outliers', 'suspectedoutliers', False],"accepted values for points 'all', 'outliers', " \
                                                                     "'suspectedoutliers'"
    assert '.' not in filename, "filename doesn't need an extension"

    figure_list = list()

    numeric_cols = col_types['numeric']

    categorical_columns = columns_with_distinct_values(df=df,
                                                       maximum_number_distinct_values=maximum_number_violinplots)

    if len(numeric_cols) != 0 and len(categorical_columns) != 0:
        for col in numeric_cols:
            for category in categorical_columns:

                plot = px.violin(data_frame=df,
                                 x=category,
                                 y=col,
                                 points=points,
                                 violinmode=violinmode,
                                 box=display_box,
                                 color=color,
                                 title='Violinplot showing distribution of {} across {} categories'.format(col,category))
                figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_box(df,
               col_types,
               points='outliers',
               boxmode='group',
               notched=False,
               color=None,
               filename='box',
               create_html=True,
               maximum_number_boxplots=7,
               orientation='v'
               ):
    """
    Creates a list of boxplot figure objects created for the entire dataset
    :param df: pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param points:str, whether to show the points in a box plot possible options are 'outliers', 'all',False,
                 'suspectedoutliers'
    :param boxmode: str, how to display the boxes in a boxplot.
                    Options are 'group' or 'overlay'. In group mode,
                    boxes are placed beside each other, in overlay mode,
                    boxes are placed on top of each other.
    :param notched: boolean, True or False, boxes are drawn with notches
    :param color:
    :param filename:
    :param create_html:boolean, whether to create an html file or not
    :param maximum_number_boxplots: int,
    :param orientation: str, how the plot should be orientated, 'v' for vertical, 'h' for horizontal
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(col_types, dict), "col_types must be a dictionary with column " \
                                        "name as keys and column type as value"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert isinstance(points, str), "points must be a string"
    assert points in ['all', 'outliers', 'suspectedoutliers', False],"accepted values for points 'all', " \
                                                                     "'outliers', " \
                                                                     "'suspectedoutliers'"
    assert isinstance(boxmode, str), "boxmode must be a string"
    assert boxmode in ['group', 'overlay'], "boxmode must be either group or overlay"
    assert isinstance(notched, bool), "notched must be a boolean"
    assert isinstance(create_html,bool), "create_html must be a boolean"
    assert isinstance(maximum_number_boxplots,int), "maximum_number_boxplots must be an integer"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(orientation,str), "orientation must be a string"
    assert orientation in ['v','h'], "options for orientation are 'v' or 'h' "

    figure_list = list()
    numeric_cols = col_types['numeric']

    categorical_columns = columns_with_distinct_values(df=df,
                                                       maximum_number_distinct_values=maximum_number_boxplots)

    if len(numeric_cols) != 0 and len(categorical_columns) != 0:

        for col in numeric_cols:
            for category in categorical_columns:

                plot = px.box(data_frame=df,
                              x=category,
                              y=col,
                              points=points,
                              boxmode=boxmode,
                              notched=notched,
                              color=color,
                              orientation=orientation,
                              title='Boxplot showing distribution of {} across {} categories'.format(col,category)
                              )
                figure_list.append(plot)

    if create_html:

        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_treemap(df,
                   col_types,
                   create_html=True,
                   filename='treemap',
                   limit=2
                   ):
    """

    :param df:pandas dataframe
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param create_html: boolean, whether to create an html file or not
    :param filename:
    :param limit:int, maximum path depth
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(create_html, bool),"create_html must be a boolean"
    assert isinstance(limit, int),"limit must be a int"
    assert limit > 1, "limit must be at least 2, to have a meaningful hierarchical chart"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert isinstance(col_types, dict), "col_types must be a dictionary"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert '.' not in filename, "filename doesn't need an extension"

    figure_list = list()

    path_list = hierarchical_path(df=df, limit=limit, col_types=col_types)

    numeric_cols = col_types['numeric']

    if len(path_list) > 1:
        for col in numeric_cols:

            plot = px.treemap(data_frame=df,
                              path=path_list,
                              values=col,
                              title='Treemap of {} across paths {}'.format(col, str(path_list))
                              )

            figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_sunburst(df,
                    col_types,
                    create_html=True,
                    filename='sunburst',
                    limit=2
                    ):
    """

    :param df:
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param create_html: boolean, whether to create an html file or not
    :param filename:
    :param limit:int, maximum path depth
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(create_html, bool),"create_html must be a boolean"
    assert isinstance(limit, int),"limit must be a int"
    assert limit > 1, "limit must be at least 2, to have a meaningful hierarchical chart"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert isinstance(col_types, dict), "col_types must be a dictionary"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert '.' not in filename, "filename doesn't need an extension"

    figure_list = list()

    path_list = hierarchical_path(df=df, limit=limit,col_types=col_types )

    numeric_cols = col_types['numeric']

    if len(path_list) > 1:
        for col in numeric_cols:

            plot = px.sunburst(data_frame=df,
                               path=path_list,
                               values=col,
                               title='Sunburst chart of {} across paths {}'.format(col, str(path_list))
                               )

            figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_correlation_plot(df,
                            correlation_method='pearson',
                            create_html=False,
                            filename='correlation_plot'):
    """
    Creates a correlation plot for the provided dataframe
    based on the correlation method supplied and returns a list of plotly graph objects
    :param df: pandas dataframe
    :param correlation_method: str, method for computing correlation, one of kendall, pearson, spearman
    :param create_html: boolean, whether to create html file or not
    :param filename:str, name of file, the extension is excluded
    :return: list of graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert isinstance(correlation_method, str), "correlation method must be a string"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert correlation_method in ['pearson', 'kendall', 'spearman']

    figure_list = list()

    if len(extract_numeric_cols(df=df))>1:

        correlation = calculate_correlation(df=df, method=correlation_method)

        plot = px.imshow(img=correlation,
                         title='Correlation plot using {} method'.format(correlation_method))

        figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_histogram(df,
                     numeric_cols,
                     nbins=20,
                     marginal=None,
                     cumulative=False,
                     histfunc='sum',
                     histnorm=None,
                     filename='histogram',
                     create_html=True,
                     orientation='v'
                     ):
    """

    :param df: pandas dataframe
    :param numeric_cols:list of columns with numeric columns
    :param nbins:int, number of bins
    :param marginal:str, marginal must be one of 'rug','box','violin','histogram
    :param cumulative:bool
    :param histfunc:str, histfunc must be one of 'count', 'sum', 'avg','min','max'
    :param histnorm:str, normalization method for histogram 'percent','probability','density','probability density'
    :param filename: str, name of file, the extension must be excluded
    :param create_html: boolean, whether to create an html file or not
    :param orientation: str, the orientation of the figure, 'v' or 'h'
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(numeric_cols, list), "numeric_cols must be a list"
    assert isinstance(nbins, int), "nbins must be an integer"
    assert marginal in ['rug','box','violin','histogram',None], "marginal options 'rug','box','violin','histogram'"
    assert isinstance(cumulative, bool), "cumulative must be a boolean"
    assert histfunc in ['count', 'sum', 'avg', 'min',
                        'max'], "histfunc must be one of 'count', 'sum', 'avg','min','max' "
    assert histnorm in ['percent', 'probability', 'density', 'probability density', None]
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert orientation in ['v','h']

    figure_list = list()

    if check_numeric(df=df,column_list=numeric_cols):
        for col in numeric_cols:
            plot = px.histogram(data_frame=df,
                                nbins=nbins,
                                x=col,
                                marginal=marginal,
                                cumulative=cumulative,
                                histfunc=histfunc,
                                histnorm=histnorm,
                                orientation=orientation,
                                title='Distribution of {}'.format(col)
                                )
            figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_scatter(df,
                   col_types,
                   filename='scatter',
                   marginal_x=None,
                   marginal_y=None,
                   log_x=False,
                   log_y=False,
                   orientation='v',
                   opacity=1.0,
                   maximum_color_groups=5,
                   create_html=True):
    """
    :param df:
    :param col_types: dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param filename:str, name of file, the extension must be excluded
    :param marginal_x:options for including marginal charts on x axis
    :param marginal_y:options for including marginal charts on y axis
    :param log_x:boolean, whether to create a log axis
    :param log_y:boolean, whether to create a log axis
    :param orientation: str, the orientation of the figure, 'v' or 'h'
    :param create_html: boolean, whether to create an html file or not
    :param opacity: float, value between 0 and 1. Sets the opacity for markers
    :param maximum_color_groups: int, maximum number of color groups in the scatter plot
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(col_types, dict), "col_types must be a dictionary with column " \
                                        "name as keys and column type as value"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert isinstance(log_y, bool)
    assert isinstance(log_x, bool)
    assert marginal_x in ['rug', 'box', 'violin', 'histogram',None], "marginal must be one of 'rug'," \
                                                                     "'box','violin','histogram'"
    assert marginal_y in ['rug', 'box', 'violin', 'histogram',None], "marginal must be one of 'rug'," \
                                                                     "'box','violin','histogram'"
    assert isinstance(opacity, float), "opacity must be a float"
    assert opacity <= 1.0, "opacity must be between 0 and 1"
    assert isinstance(create_html, bool), "create_html must be a boolean"
    assert orientation in ['v', 'h']
    assert isinstance(maximum_color_groups, int)

    figure_list = list()
    numeric_cols = col_types['numeric']

    if len(numeric_cols) < 2:
        # scatter plots need at least 2 columns of numeric type
        return figure_list
    plot_pairs = create_pairs(df, numeric_cols=numeric_cols)

    columns_with_distinct = columns_with_distinct_values(df,
                                                         maximum_number_distinct_values=maximum_color_groups,
                                                         categorical_only=True)
    if len(columns_with_distinct) == 0:
        columns_with_distinct.append(None)

    for pair in plot_pairs:
        x,y = pair
        for color in columns_with_distinct:
            plot = px.scatter(data_frame=df,
                              x=x,
                              y=y,
                              log_y=log_y,
                              log_x=log_x,
                              opacity=opacity,
                              orientation=orientation,
                              marginal_x=marginal_x,
                              marginal_y=marginal_y,
                              color=color,
                              title='Plot of {} vs {} with color {}'.format(x,y,color))
            figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list


def create_density_chart(df,
                         col_types,
                         nbinsx=None,
                         nbinsy=None,
                         maximum_color_groups=4,
                         typr_of_chart='contour',
                         filename='density_charts',
                         orientation='v',
                         marginal_x=None,
                         marginal_y=None,
                         log_x=False,
                         log_y=False,
                         histfunc='sum',
                         histnorm=None,
                         create_html=True
                         ):
    """
    Create density chart and density heatmap charts
    :param df: pandas dataframe
    :param col_types:  dictionary that contains mapping of column type to list of column names
                    It is the output of extract_col_types in tabular module
    :param nbinsx: int, number of bins on x axis
    :param nbinsy: int, number of bins on y axis
    :param maximum_color_groups: int, maximum number of color groups in the density charts
    :param typr_of_chart: str, 'contour' or 'heatmap'
    :param filename: str, name of file, the extension must be excluded
    :param orientation: str, the orientation of the figure, 'v' or 'h'
    :param marginal_x: options for including marginal charts on x axis, one of 'rug', 'box', 'violin', 'histogram',None
    :param marginal_y: options for including marginal charts on y axis, one of 'rug', 'box', 'violin', 'histogram',None
    :param log_x: boolean, whether to create a log x axis
    :param log_y: boolean, whether to create a log y axis
    :param histfunc: histfunc must be one of 'count', 'sum', 'avg','min','max'
    :param histnorm: str, normalization method for histogram 'percent','probability','density','probability density'
    :param create_html: boolean, whether to create an html file or not
    :return:list which contains plotly graph objects
    """
    assert isinstance(df, pd.DataFrame), "df must be a pandas dataframe"
    assert isinstance(col_types, dict), "col_types must be a dictionary with column " \
                                        "name as keys and column type as value"
    keys = col_types.keys()
    assert 'numeric' in keys, "col_types dictionary missing key numeric"
    assert 'object' in keys, "col_types dictionary missing key object"
    assert 'boolean' in keys, "col_types dictionary missing key boolean"
    assert 'categorical' in keys, "col_types dictionary missing key categorical"
    assert 'datetime' in keys, "col_types dictionary missing key datetime"
    assert 'timedelta' in keys, "col_types dictionary missing key timedelta"
    assert isinstance(nbinsx,int) or nbinsx is None
    assert isinstance(nbinsy, int) or nbinsy is None
    assert isinstance(log_y, bool)
    assert isinstance(log_x, bool)
    assert marginal_x in ['rug', 'box', 'violin', 'histogram',None], "marginal must be one of 'rug'," \
                                                                     "'box','violin','histogram'"
    assert marginal_y in ['rug', 'box', 'violin', 'histogram',None], "marginal must be one of 'rug'," \
                                                                     "'box','violin','histogram'"
    assert histfunc in ['count', 'sum', 'avg', 'min',
                        'max'], "histfunc must be one of 'count', 'sum', 'avg','min','max' "
    assert histnorm in ['percent', 'probability', 'density', 'probability density', None]
    assert orientation in ['v', 'h']
    assert isinstance(filename, str), "filename must be a string with a dot or an extension"
    assert '.' not in filename, "filename doesn't need an extension"
    assert typr_of_chart in ['contour','heatmap']

    figure_list = list()
    numeric_cols = col_types['numeric']

    if len(numeric_cols) < 2:
        # density contour at least 2 columns of numeric type
        return figure_list

    # create pairing for x and y values
    plot_pairs = create_pairs(df, numeric_cols=numeric_cols)

    columns_with_distinct = columns_with_distinct_values(df,
                                                         maximum_number_distinct_values=maximum_color_groups,
                                                         categorical_only=True)

    if len(columns_with_distinct) == 0:
        columns_with_distinct.append(None)

    for pair in plot_pairs:
        x,y = pair
        for z in numeric_cols:
            if typr_of_chart == 'contour':
                for color in columns_with_distinct:
                    title = "Density contour of {} vs {},with {} on z axis".format(x, y, z)
                    plot = px.density_contour(data_frame=df,
                                              x=x,
                                              y=y,
                                              z=z,
                                              color=color,
                                              histnorm=histnorm,
                                              histfunc=histfunc,
                                              nbinsx=nbinsx,
                                              nbinsy=nbinsy,
                                              log_y=log_y,
                                              log_x=log_x,
                                              marginal_x=marginal_x,
                                              marginal_y=marginal_y,
                                              orientation=orientation,
                                              title=title)
                    figure_list.append(plot)

            if typr_of_chart == 'heatmap':
                title = "Density heatmap of {} vs {},with {} on z axis".format(x, y, z)

                plot = px.density_heatmap(data_frame=df,
                                          x=x,
                                          y=y,
                                          z=z,
                                          histnorm=histnorm,
                                          histfunc=histfunc,
                                          nbinsx=nbinsx,
                                          nbinsy=nbinsy,
                                          log_y=log_y,
                                          log_x=log_x,
                                          marginal_x=marginal_x,
                                          marginal_y=marginal_y,
                                          orientation=orientation,
                                          title=title)
                figure_list.append(plot)

    if create_html:
        figures_to_html(figs=figure_list, filename=filename)

    return figure_list




















from sklearn.base import BaseEstimator, TransformerMixin
from .utils import check_columns, check_dataframe, check_numeric


class ColumnAdder(BaseEstimator, TransformerMixin):
    """
    Performs addition of columns
    """
    def __init__(self, column_list, new_column_name):
        """
        Adds columns together in a dataframe and creates a new column

        :param column_list: list of columns to be added together, they must all be numeric columns
        :param new_column_name:str, name of new column created from adding the columns in column list together

        Usage
        >>> import pandas as pd
        >>> from datamallet.tabular.feature import ColumnAdder
        >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'C':['dog','cat', 'sheep','dog','cat'],'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

        >>> column_adder = ColumnAdder(column_list=['A','B'],new_column_name='Z')
        >>> added_df = column_adder.transform(X=df)
        >>> print(added_df)

           A   B      C       D      E   Z
        0  1   2    dog    male   True   3
        1  2   4    cat    male   True   6
        2  3   6  sheep    male  False   9
        3  4   8    dog  female   True  12
        4  5  10    cat  female   True  15


        """
        self.column_list = column_list
        self.new_column_name = new_column_name
        assert isinstance(column_list,list), "column_list must be a list"
        assert isinstance(new_column_name, str), "new_column_name must be a string"

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if check_dataframe(X) \
                and check_numeric(df=X, column_list=self.column_list):
            X = X.copy()
            X[self.new_column_name] = X.loc[:, self.column_list].sum(axis=1)

        return X


class ColumnMultiplier(BaseEstimator,TransformerMixin):
    def __init__(self, column_list, new_column_name):
        """
        Adds columns together and creates a new column

        :param column_list:list of columns to be multiplied together, they must all be numeric columns
        :param new_column_name:str, name of new column created from adding the columns in column list together

        Usage
        >>> import pandas as pd
        >>> from datamallet.tabular.feature import ColumnMultiplier
        >>> df3 = pd.DataFrame({'A':[1,1,2,1,1],'B':[2,2,1,2,0],})
        >>> multiplied_df = ColumnMultiplier(column_list=['A','B'], new_column_name='Z').transform(X=df3)
        >>> print(multiplied_df)
           A  B  Z
        0  1  2  2
        1  1  2  2
        2  2  1  2
        3  1  2  2
        4  1  0  0


        """
        self.column_list = column_list
        self.new_column_name = new_column_name
        assert isinstance(column_list, list),"column_list must be a list"
        assert isinstance(new_column_name, str), "new_column_name must be a string"

    def fit(self, X, y=None):
        return self

    def transform(self,X, y=None):
        if check_dataframe(X) and \
                 check_numeric(df=X, column_list=self.column_list):
            X = X.copy()
            X[self.new_column_name] = X.loc[:, self.column_list].prod(axis=1,
                                                                      numeric_only=True,
                                                                      skipna=True)
        return X


class ColumnSubtraction(BaseEstimator, TransformerMixin):
    def __init__(self, left, right, new_column_name):
        """

        :param left: name of column on the left side of the minus side
        :param right: name of column on the right side of the minus side
        :param new_column_name:str, name of new column

        Usage
        >>> import pandas as pd
        >>> from datamallet.tabular.feature import ColumnSubtraction
        >>> df3 = pd.DataFrame({'A':[1,1,2,1,1],'B':[2,2,1,2,0],})
        >>> subtracted_df = ColumnSubtraction(left='A', right='B', new_column_name='C').transform(df3)
        >>> print(subtracted_df)
           A  B  C
        0  1  2 -1
        1  1  2 -1
        2  2  1  1
        3  1  2 -1
        4  1  0  1

        """
        self.left = left
        self.right = right
        self.new_column_name = new_column_name
        assert isinstance(left,str), "left must be a string with name of column"
        assert isinstance(right, str), "right must be a string with name of column"
        assert isinstance(new_column_name, str), "new_column_name must be a string"

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if check_dataframe(X) and \
                check_numeric(df=X, column_list=[self.left,self.right]):
            X = X.copy()
            X[self.new_column_name] = X.loc[:, self.left] - X.loc[:, self.right]
        return X


class ExpandingTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, column_list ,min_periods=1, aggregation_function='sum'):
        """
        Provides functionality for carrying out expanding aggregation
        :param column_list: python list containing column names for which the transformation should be performed on
        :param min_periods: int, minimum number of observations in window required to have a value
        :param aggregation_function: str, aggregation function to be applied, options are 'mean'
               , 'max', 'std' , 'sum', 'min'

        Usage
        >>> import pandas as pd
        >>> from datamallet.tabular.feature import ExpandingTransformer
        >>> df = pd.DataFrame({"B": [0, 1, 2, np.nan, 4], "A": [0, 1, 0, np.nan, 4]})
        >>> expander = ExpandingTransformer(column_list=['B'],min_periods=2,aggregation_function='sum')
        >>> df_new = expander.transform(X=df)
        >>> print(df_new)
             B    A
        0  NaN  0.0
        1  1.0  1.0
        2  3.0  0.0
        3  3.0  NaN
        4  7.0  4.0

        """

        self.column_list = column_list
        self.min_periods = min_periods
        self.aggregation_function = aggregation_function
        assert isinstance(column_list, list)
        assert isinstance(min_periods, int)
        assert aggregation_function in ['mean','max','std','sum','min']

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if check_columns(column_list=self.column_list, df=X) and check_dataframe(X):
            X = X.copy()
            for col in self.column_list:
                X[col] = X[col].expanding(self.min_periods, axis=0).agg(self.aggregation_function)

        return X


class GroupbyTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, column_list, aggregation_method='mean'):
        """

        :param column_list:list of column names to be used for the aggregation (groupin)
        :param aggregation_method:str, aggregation method, one of 'sum','mean','std'

        Usage
        >>> import pandas as pd
        >>> from datamallet.tabular.feature import GroupbyTransformer
        >>> df4 = pd.DataFrame({'Age':[1,2,3,4,5],
                   'Gender':['Male','Female','Unknown','Male','Female'],
                    'City':['austin', 'austin', 'lagos','abuja', 'ibadan']
                   })
        >>> grouped_df = GroupbyTransformer(column_list=['Gender', 'City'], aggregation_method='mean').transform(df4)

        >>> print(grouped_df)

                        Age
        Gender  City
        Female  austin    2
                ibadan    5
        Male    abuja     4
                austin    1
        Unknown lagos     3

        >>> grouped_df = GroupbyTransformer(column_list=['Gender'], aggregation_method='sum').transform(df4)

        >>> print(grouped_df)

                 Age
        Gender
        Female     7
        Male       5
        Unknown    3


        """
        self.column_list = column_list
        self.aggregation_method = aggregation_method
        assert isinstance(column_list, list)
        assert aggregation_method in ['mean', 'std', 'sum']

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if check_dataframe(df=X) and check_columns(df=X, column_list=self.column_list):
            X = X.copy()
            X = X.groupby(self.column_list).agg(self.aggregation_method)

            return X





from .utils import check_dataframe, time_index
from sklearn.base import BaseEstimator, TransformerMixin


class Resampler(BaseEstimator,TransformerMixin):
    def __init__(self, rule, aggregation_method='mean'):
        """
        Resampler is a transformer for changing the frequency of the data.
        :param rule: str, or date offset representing the target resolution
                    1D represents a day
                    1H represents an hour
                    1T represents a minute
                    1S represents a second
                    1L represents a millisecond
                    1U represents a microsecond
                    1N represents a nanosecond

                    different numbers can be placed as prefix eg 2H is 2 hours frequency etc

                    more info here https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects

        :param aggregation_method:str, aggregation method, one of mean, std, max, min, sum

        Usage
        >>> import pandas as pd
        >>> from datamallet.tabular.timeseries import Resampler
        >>> price = [10, 11, 9, 13, 14, 18, 17, 19,10, 11, 9, 13, 14, 18, 17, 19,50, 60, 40, 100, 50, 100, 40,10, 11, 9, 13, 14, 18, 17, 19,10,]
        >>> volume = [50, 60, 40, 100, 50, 100, 40, 50,10, 11, 9, 13, 14, 18, 17, 19,50, 60, 40, 100, 50, 100, 40,10, 11, 9, 13, 14, 18, 17, 19,10,]
        >>> d = {'price':price,'volume':volume}
        >>> df = pd.DataFrame(d)
        >>> df['week_starting'] = pd.date_range('01/01/2018',periods=32, freq='1H')
        >>> df.index = df['week_starting']
        >>> df.drop('week_starting', inplace=True, axis=1)

        >>> resampler = Resampler(rule='24H', aggregation_method='mean')
        >>> df_r = resampler.transform(X=df)
        >>> print(df_r)
                        price     volume
        week_starting
        2018-01-01     28.000  43.791667
        2018-01-02     13.875  13.875000

        """
        self.rule = rule
        # self.on = on
        self.aggregation_method = aggregation_method
        assert isinstance(rule, str), "rule must be a string"
        assert aggregation_method in ['mean','std','min','max','sum']

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X = X.copy()
        if check_dataframe(X) and time_index(X):
            resampled_x = X.resample(rule=self.rule).agg(self.aggregation_method)
            return resampled_x

        else:
            return X



import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from .utils import (check_columns,
                    extract_numeric_cols,
                    check_dictionary,
                    check_dataframe,
                    column_mean,percentage_missing,
                    check_numeric)


class ColumnDropper(BaseEstimator, TransformerMixin):
    def __init__(self,
                 column_list):
        """
        This class drops columns from a dataframe, this operation is done inplace.

        :param column_list: list which contains the names of columns in the dataframe to be dropped

        Usage
        >>> from datamallet.tabular.preprocess import ColumnDropper
        >>> import pandas as pd
        >>> df = pd.DataFrame({'A':[1,2,3,4,5],
                   'B':[2,4,6,8,10],
                   'C':['dog','cat', 'sheep','dog','cat'],
                   'D':['male','male','male','female','female'],
                   'E':[True,True,False,True,True]})

        # initialize the ColumnDropper class
        >>> column_dropper = ColumnDropper(column_list=['C','D','E'])

        # call the transform method
        >>> dropped_df = column_dropper.transform(X=df)

        # check
        >>> print(dropped_df.columns)
        Index(['A', 'B'], dtype='object')

        """
        self.column_list = column_list

    def fit(self, X, y=None):
        assert (isinstance(X, pd.DataFrame)), 'X needs to be a pandas dataframe'
        return self

    def transform(self, X, y=None):
        if check_columns(X, self.column_list) and check_dataframe(X):
            X = X.copy()
            X.drop(labels=self.column_list, inplace=True, axis=1)
        return X


class NaFiller(BaseEstimator, TransformerMixin):
    def __init__(self,column_list,
                 method='mean',
                 limit=None,
                 ):
        """
          Transformer for filling missing values using various methods or values

          :param value: float, int or dictionary which contains values to be used for filling values
                          , only provide if method is None
          :param method: str, method for use with filling missing value, current values allowed are
                          bfill for backfilling
                          ffill for forward filling
                          mean for filling with the average values for each column
          :param limit:int, default = None, the maximum number of missing values to be filled
          :param column_list:, list, default is None, list of column names to apply the imputation to

          Usage
          # using method = bfill
          >>> import pandas as pd
          >>> import numpy as np
          >>> from datamallet.tabular.preprocess import NaFiller
          >>> df2 = pd.DataFrame({'A':[np.nan,2,3,4,5,8],'B':[2,np.nan,np.nan,np.nan,10,9],'C':[1,3,5,np.nan,np.nan,7]})
          >>> print(df2)
               A     B    C
            0  NaN   2.0  1.0
            1  2.0   NaN  3.0
            2  3.0   NaN  5.0
            3  4.0   NaN  NaN
            4  5.0  10.0  NaN
            5  8.0   9.0  7.0
          >>> nafiller = NaFiller( method='bfill', limit=None, column_list=None)
          >>> cx = nafiller.transform(X=df2)
          >>> print(cx)
               A     B    C
            0  2.0   2.0  1.0
            1  2.0  10.0  3.0
            2  3.0  10.0  5.0
            3  4.0  10.0  7.0
            4  5.0  10.0  7.0
            5  8.0   9.0  7.0

            # Using ffill
          >>> nafiller = NaFiller( method='ffill', limit=None, column_list=None)
          >>> cx = nafiller.transform(X=df2)
          >>> print(cx)
               A     B    C
            0  NaN   2.0  1.0
            1  2.0   2.0  3.0
            2  3.0   2.0  5.0
            3  4.0   2.0  5.0
            4  5.0  10.0  5.0
            5  8.0   9.0  7.0

          >>> nafiller = NaFiller( method='mean', limit=None, column_list=None)
          >>> cx = nafiller.transform(X=df2)
          >>> print(cx)
              A     B    C
          0  4.4   2.0  1.0
          1  2.0   7.0  3.0
          2  3.0   7.0  5.0
          3  4.0   7.0  4.0
          4  5.0  10.0  4.0
          5  8.0   9.0  7.0

          >>> d = pd.DataFrame({'A':[np.nan,2,3,4,5,8],'B':[2,np.nan,np.nan,np.nan,10,9],'C':[1,3,5, np.nan, np.nan,7],'D':[4,np.nan,np.nan, np.nan, np.nan,7]})
          >>> nafiller = NaFiller( method='mean', column_list=['B','C'], limit=1)
          >>> cx = nafiller.transform(X=d)
               A     B    C    D
           0  NaN   2.0  1.0  4.0
           1  2.0   7.0  3.0  NaN
           2  3.0   NaN  5.0  NaN
           3  4.0   NaN  4.0  NaN
           4  5.0  10.0  NaN  NaN
           5  8.0   9.0  7.0  7.0


        """

        self.method = method
        self.limit = limit
        self.column_list = column_list
        assert method in ['bfill', 'ffill','mean']
        assert isinstance(limit, int) or limit is None
        assert isinstance(column_list,list) or column_list is None

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        assert isinstance(X, pd.DataFrame)
        X = X.copy()

        if self.method == 'bfill':
            if isinstance(self.column_list, list) and check_columns(df=X,column_list=self.column_list):
                for col in self.column_list:
                    X[col] = X[col].fillna(method='bfill', limit=self.limit)
                return X
            elif self.column_list is None:
                X.fillna(method='bfill', inplace=True, limit=self.limit)
                return X
            else:
                return X

        if self.method == 'ffill':
            if isinstance(self.column_list, list) and check_columns(df=X,column_list=self.column_list):
                for col in self.column_list:
                    X[col] = X[col].fillna(method='ffill', limit=self.limit)
                return X
            elif self.column_list is None:
                X.fillna(method='ffill', inplace=True, limit=self.limit)
                return X
            else:
                return X

        if self.method == 'mean':

            if isinstance(self.column_list, list) and check_numeric(df=X,column_list=self.column_list):
                X.fillna(value=column_mean(df=X, column_list=self.column_list), inplace=True,limit=self.limit)
                return X
            if self.column_list is None:
                numeric_cols = extract_numeric_cols(df=X)

                # if no columns are specified, use only numeric columns
                X.fillna(value=column_mean(df=X, column_list=numeric_cols), inplace=True, limit=self.limit)
                return X


class ConstantValueFiller(BaseEstimator, TransformerMixin):
    def __init__(self, value, limit=None):
        """
        Performs Missing value imputation using a constant value or dictionary,
        this dictionary has the column name as key and the value to use to fill as value
        :param value:  dictionary which maps column name to value or interger or string
        :param limit:int, (default = None) the maximum number of missing value per column to be filled,
                if None, all missing values would be filled

        Usage
        >>> import pandas as pd
        >>> import numpy as np
        >>> df2 = pd.DataFrame({'A':[np.nan,2,3,4,5,8],'B':[2,np.nan,np.nan,np.nan,10,9],'C':[1,3,5, np.nan, np.nan,7]})
        >>> cvf = ConstantValueFiller(value={'A':100, 'B':200,'C':300}, limit=1)
        >>> vx = cvf.transform(X=df2)
        >>> print(vx)
               A      B      C
        0  100.0    2.0    1.0
        1    2.0  200.0    3.0
        2    3.0    NaN    5.0
        3    4.0    NaN  300.0
        4    5.0   10.0    NaN
        5    8.0    9.0    7.0

        >>> cvf = ConstantValueFiller(value=200, limit=None)
        >>> vx = cvf.transform(X=df2)
        >>> print(vx)
               A      B      C
        0  200.0    2.0    1.0
        1    2.0  200.0    3.0
        2    3.0  200.0    5.0
        3    4.0  200.0  200.0
        4    5.0   10.0  200.0
        5    8.0    9.0    7.0
        """
        self.value = value
        self.limit = limit
        assert isinstance(value, dict) or value is not None, "fill_dict is expected to be a dictionary"
        assert isinstance(limit, int) or limit is None

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        X = X.copy()

        if check_dataframe(df=X):
            if isinstance(self.value, int) or isinstance(self.value, float) or isinstance(self.value, str):
                X.fillna(value=self.value, inplace=True, limit=self.limit)
                return X
            if check_dictionary(df=X, column_dict=self.value):
                X.fillna(value=self.value, inplace=True, limit=self.limit)
                return X

        return X


class ColumnRename(BaseEstimator, TransformerMixin):
    def __init__(self, rename_dictionary):
        """
        Rename columns in place
        :param rename_dictionary: python dictionary with the old names as keys and new names as value.

        Usage
        >>> from datamallet.tabular.preprocess import ColumnRename
        >>> import pandas as pd
        >>> df = pd.DataFrame({'A':[1,2,3,4,5],'B':[2,4,6,8,10],'D':['male','male','male','female','female'],'E':[True,True,False,True,True]})

        >>> rename_dictionary={'A':'V','B':'W'}

        >>> columnrenamer = ColumnRename(rename_dictionary=rename_dictionary)
        >>> renamed_df = columnrenamer.transform(X=df)

        >>> print(renamed_df.columns)
        Index(['V', 'W', 'C', 'D', 'E'], dtype='object')

        """
        self.rename_dictionary = rename_dictionary

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        assert isinstance(self.rename_dictionary, dict), 'Rename dictionary must be a dictionary'

        col_list = list()
        for old_name, new_name in self.rename_dictionary.items():
            col_list.append(old_name)

        if check_columns(df=X, column_list=col_list) and check_dataframe(df=X):
            X = X.copy()
            X.rename(columns=self.rename_dictionary, inplace=True)

        return X


class NADropper(BaseEstimator, TransformerMixin):
    def __init__(self, axis=1,how='all',thresh=None):
        """
        Transformer for dropping columns with missing value.
        :param axis: 0 or 'index', drops rows which contain missing value, 1 or 'columns' drops columns which contain missing value.
        :param how: str, 'any','all'. drop row or column if any NA is present, drop row or column if all values are missing.
        :param thresh: int, require that many non NA values

        Usage
        >>> import pandas as pd
        >>> import numpy as np
        >>> df3 = pd.DataFrame({'A':[np.nan,2,3,4,5],'B':[2,np.nan,np.nan,np.nan,10],
            ... 'C':['dog','cat', 'sheep','dog','cat'],'D':['male','male','male','female','female'],
            ... 'E':[True,True,False,True,True]})
        >>> nadropper = NADropper(axis='columns', how='any')
        >>> droped2 = nadropper.transform(X=df2)
               C       D      E
        0    dog    male   True
        1    cat    male   True
        2  sheep    male  False
        3    dog  female   True
        4    cat  female   True

        """
        self.axis = axis
        self.how = how
        self.thresh = thresh
        assert isinstance(thresh,int) or thresh is None,"thresh is expected to be an integer"
        assert how in ['any','all']
        assert axis in [0,1,'index','columns'],""

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        if check_dataframe(df=X):
            X = X.copy()
            X.dropna(axis=self.axis, how=self.how, thresh=self.thresh, inplace=True)

        return X


class DropPercentageMissing(BaseEstimator, TransformerMixin):
    def __init__(self, threshold=50):
        """
        Drops column which have a percentage of missing value greater than or equal to the threshold
        :param threshold: int

        Usage
         >>> import pandas as pd
        >>> from datamallet.tabular.preprocess import DropPercentageMissing
        >>> df = pd.DataFrame(dict(age=[5, 6, np.NaN],born=[pd.NaT, pd.Timestamp('1939-05-27'),pd.Timestamp('1940-04-25')],
        ... name=['Alfred', 'Batman', ''],toy=[None, 'Batmobile', 'Joker']))
        >>> bn = DropPercentageMissing(threshold=20).transform(X=df)
        >>> print(bn)
             name
        0  Alfred
        1  Batman
        2

        """
        assert isinstance(threshold, int)
        self.threshold = threshold

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        assert isinstance(X, pd.DataFrame)
        X = X.copy()
        percentage_mis_dict = percentage_missing(df=X)
        drop_list = list()

        for column, percentage in percentage_mis_dict.items():
            if percentage >= self.threshold:
                drop_list.append(column)

        X.drop(axis='columns', inplace=True,labels=drop_list)

        return X













